# Copyright 2022 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import hashlib
import io
import os
from unittest import mock

import socket
import sys
import tempfile
import time
import uuid

from keystoneauth1 import exceptions as keystone_exc
from os_brick.initiator import connector
from oslo_concurrency import processutils
from oslo_utils.secretutils import md5
from oslo_utils import units

from glance_store.common import attachment_state_manager
from glance_store.common import cinder_utils
from glance_store import exceptions
from glance_store import location

sys.modules['glance_store.common.fs_mount'] = mock.Mock()
from glance_store._drivers.cinder import store as cinder # noqa
from glance_store._drivers.cinder import nfs # noqa


class TestCinderStoreBase(object):

    def test_get_cinderclient(self):
        cc = self.store.get_cinderclient(self.context)
        self.assertEqual('fake_token', cc.client.auth.token)
        self.assertEqual('http://foo/public_url', cc.client.auth.endpoint)

    def _get_cinderclient_with_user_overriden(self, group='glance_store',
                                              **kwargs):

        cinderclient_opts = {
            'cinder_store_user_name': 'test_user',
            'cinder_store_password': 'test_password',
            'cinder_store_project_name': 'test_project',
            'cinder_store_auth_address': 'test_address'}

        cinderclient_opts.update(kwargs)
        self.config(**cinderclient_opts, group=group)
        cc = self.store.get_cinderclient(self.context)
        return cc

    def _test_get_cinderclient_with_user_overriden(
            self, group='glance_store'):
        cc = self._get_cinderclient_with_user_overriden(
            group=group)
        self.assertEqual('test_project',
                         cc.client.session.auth.project_name)
        self.assertEqual('Default',
                         cc.client.session.auth.project_domain_name)

    def _test_get_cinderclient_with_user_overriden_and_region(
            self, group='glance_store'):
        cc = self._get_cinderclient_with_user_overriden(
            group=group, **{'cinder_os_region_name': 'test_region'})
        self.assertEqual('test_region', cc.client.region_name)

    def _test_get_cinderclient_with_api_insecure(
            self, group='glance_store'):
        self.config()
        with mock.patch.object(
            cinder.ksa_session, 'Session') as fake_session, \
            mock.patch.object(
                cinder.ksa_identity, 'V3Password') as fake_auth_method:
            fake_auth = fake_auth_method()
            self._get_cinderclient_with_user_overriden(
                group=group, **{'cinder_api_insecure': True})
            fake_session.assert_called_once_with(auth=fake_auth, verify=False)

    def _test_get_cinderclient_with_ca_certificates(
            self, group='glance_store'):
        fake_cert_path = 'fake_cert_path'
        with mock.patch.object(
            cinder.ksa_session, 'Session') as fake_session, \
            mock.patch.object(
                cinder.ksa_identity, 'V3Password') as fake_auth_method:
            fake_auth = fake_auth_method()
            self._get_cinderclient_with_user_overriden(
                group=group, **{'cinder_ca_certificates_file': fake_cert_path})
            fake_session.assert_called_once_with(
                auth=fake_auth, verify=fake_cert_path)

    def _test_get_cinderclient_cinder_endpoint_template(self,
                                                        group='glance_store'):
        fake_endpoint = 'http://cinder.openstack.example.com/v2/fake_project'
        self.config(cinder_endpoint_template=fake_endpoint, group=group)
        with mock.patch.object(
                cinder.ksa_token_endpoint, 'Token') as fake_token:
            self.store.get_cinderclient(self.context)
            fake_token.assert_called_once_with(endpoint=fake_endpoint,
                                               token=self.context.auth_token)

    def test_get_cinderclient_endpoint_exception(self):
        with mock.patch.object(cinder.ksa_session, 'Session'), \
            mock.patch.object(cinder.ksa_identity, 'V3Password'), \
            mock.patch.object(
                cinder.Store, 'is_user_overriden', return_value=False), \
            mock.patch.object(
                cinder.keystone_sc, 'ServiceCatalogV2') as service_catalog:
            service_catalog.side_effect = keystone_exc.EndpointNotFound
            self.assertRaises(
                exceptions.BadStoreConfiguration, self.store.get_cinderclient,
                self.context)

    def test__check_context(self):
        with mock.patch.object(cinder.Store, 'is_user_overriden',
                               return_value=True) as fake_overriden:
            self.store._check_context(self.context)
            fake_overriden.assert_called_once()

    def test__check_context_no_context(self):
        with mock.patch.object(
                cinder.Store, 'is_user_overriden', return_value=False):
            self.assertRaises(
                exceptions.BadStoreConfiguration, self.store._check_context,
                None)

    def test__check_context_no_service_catalog(self):
        with mock.patch.object(
                cinder.Store, 'is_user_overriden', return_value=False):
            fake_context = mock.MagicMock(service_catalog=None)
            self.assertRaises(
                exceptions.BadStoreConfiguration, self.store._check_context,
                fake_context)

    def test_temporary_chown(self):
        fake_stat = mock.MagicMock(st_uid=1)

        with mock.patch.object(os, 'stat', return_value=fake_stat), \
                mock.patch.object(os, 'getuid', return_value=2), \
                mock.patch.object(processutils, 'execute') as mock_execute, \
                mock.patch.object(cinder.Store, 'get_root_helper',
                                  return_value='sudo'):
            with self.store.temporary_chown('test'):
                pass
            expected_calls = [mock.call('chown', 2, 'test', run_as_root=True,
                                        root_helper='sudo'),
                              mock.call('chown', 1, 'test', run_as_root=True,
                                        root_helper='sudo')]
            self.assertEqual(expected_calls, mock_execute.call_args_list)

    @mock.patch.object(time, 'sleep')
    def test_wait_volume_status(self, mock_sleep):
        fake_manager = mock.MagicMock(get=mock.Mock())
        volume_available = mock.MagicMock(manager=fake_manager,
                                          id='fake-id',
                                          status='available')
        volume_in_use = mock.MagicMock(manager=fake_manager,
                                       id='fake-id',
                                       status='in-use')
        fake_manager.get.side_effect = [volume_available, volume_in_use]
        self.assertEqual(volume_in_use,
                         self.store._wait_volume_status(
                             volume_available, 'available', 'in-use'))
        fake_manager.get.assert_called_with('fake-id')
        mock_sleep.assert_called_once_with(0.5)

    @mock.patch.object(time, 'sleep')
    def test_wait_volume_status_unexpected(self, mock_sleep):
        fake_manager = mock.MagicMock(get=mock.Mock())
        volume_available = mock.MagicMock(manager=fake_manager,
                                          id='fake-id',
                                          status='error')
        fake_manager.get.return_value = volume_available
        self.assertRaises(exceptions.BackendException,
                          self.store._wait_volume_status,
                          volume_available, 'available', 'in-use')
        fake_manager.get.assert_called_with('fake-id')

    @mock.patch.object(time, 'sleep')
    def test_wait_volume_status_timeout(self, mock_sleep):
        fake_manager = mock.MagicMock(get=mock.Mock())
        volume_available = mock.MagicMock(manager=fake_manager,
                                          id='fake-id',
                                          status='available')
        fake_manager.get.return_value = volume_available
        self.assertRaises(exceptions.BackendException,
                          self.store._wait_volume_status,
                          volume_available, 'available', 'in-use')
        fake_manager.get.assert_called_with('fake-id')

    def _test_open_cinder_volume(self, open_mode, attach_mode, error,
                                 multipath_supported=False,
                                 enforce_multipath=False,
                                 encrypted_nfs=False, qcow2_vol=False,
                                 multiattach=False,
                                 update_attachment_error=None):
        fake_volume = mock.MagicMock(id=str(uuid.uuid4()), status='available',
                                     multiattach=multiattach)
        fake_volume.manager.get.return_value = fake_volume
        fake_attachment_id = str(uuid.uuid4())
        fake_attachment_create = {'id': fake_attachment_id}
        if encrypted_nfs or qcow2_vol:
            fake_attachment_update = mock.MagicMock(
                id=fake_attachment_id,
                connection_info={'driver_volume_type': 'nfs'})
        else:
            fake_attachment_update = mock.MagicMock(
                id=fake_attachment_id,
                connection_info={'driver_volume_type': 'fake'})
        fake_conn_info = mock.MagicMock(connector={})
        fake_volumes = mock.MagicMock(get=lambda id: fake_volume)
        fake_client = mock.MagicMock(volumes=fake_volumes)
        _, fake_dev_path = tempfile.mkstemp(dir=self.test_dir)
        fake_devinfo = {'path': fake_dev_path}
        fake_connector = mock.MagicMock(
            connect_volume=mock.Mock(return_value=fake_devinfo),
            disconnect_volume=mock.Mock())

        @contextlib.contextmanager
        def fake_chown(path):
            yield

        def do_open():
            if multiattach:
                with mock.patch.object(
                        attachment_state_manager._AttachmentStateManager,
                        'get_state') as mock_get_state:
                    mock_get_state.return_value.__enter__.return_value = (
                        attachment_state_manager._AttachmentState())
                    with self.store._open_cinder_volume(
                            fake_client, fake_volume, open_mode):
                        pass
            else:
                with self.store._open_cinder_volume(
                        fake_client, fake_volume, open_mode):
                    if error:
                        raise error

        def fake_factory(protocol, root_helper, **kwargs):
            return fake_connector

        root_helper = "sudo glance-rootwrap /etc/glance/rootwrap.conf"
        with mock.patch.object(cinder.Store,
                               '_wait_volume_status',
                               return_value=fake_volume), \
                mock.patch.object(cinder.Store, 'temporary_chown',
                                  side_effect=fake_chown), \
                mock.patch.object(cinder.Store, 'get_root_helper',
                                  return_value=root_helper), \
                mock.patch.object(connector.InitiatorConnector, 'factory',
                                  side_effect=fake_factory
                                  ) as fake_conn_obj, \
                mock.patch.object(cinder_utils.API,
                                  'attachment_create',
                                  return_value=fake_attachment_create
                                  ) as attach_create, \
                mock.patch.object(cinder_utils.API,
                                  'attachment_update',
                                  return_value=fake_attachment_update
                                  ) as attach_update, \
                mock.patch.object(cinder_utils.API,
                                  'attachment_delete') as attach_delete, \
                mock.patch.object(cinder_utils.API,
                                  'attachment_get') as attach_get, \
                mock.patch.object(cinder_utils.API,
                                  'attachment_complete') as attach_complete, \
                mock.patch.object(socket,
                                  'gethostname') as mock_get_host, \
                mock.patch.object(socket,
                                  'getaddrinfo') as mock_get_host_ip, \
                mock.patch.object(cinder.strutils, 'mask_dict_password'):

            if update_attachment_error:
                attach_update.side_effect = update_attachment_error

            fake_host = 'fake_host'
            fake_addr_info = [[0, 1, 2, 3, ['127.0.0.1']]]
            fake_ip = fake_addr_info[0][4][0]
            mock_get_host.return_value = fake_host
            mock_get_host_ip.return_value = fake_addr_info

            with mock.patch.object(connector,
                                   'get_connector_properties',
                                   return_value=fake_conn_info) as mock_conn:
                if error:
                    self.assertRaises(error, do_open)
                elif encrypted_nfs or qcow2_vol:
                    fake_volume.encrypted = False
                    if encrypted_nfs:
                        fake_volume.encrypted = True
                    elif qcow2_vol:
                        attach_get.return_value = mock.MagicMock(
                            connection_info={'format': 'qcow2'})
                    try:
                        with self.store._open_cinder_volume(
                                fake_client, fake_volume, open_mode):
                            pass
                    except exceptions.BackendException:
                        attach_delete.assert_called_once_with(
                            fake_client, fake_attachment_id)
                elif update_attachment_error:
                    self.assertRaises(type(update_attachment_error), do_open)
                else:
                    do_open()
                if update_attachment_error:
                    attach_delete.assert_called_once_with(
                        fake_client, fake_attachment_id)
                elif not (encrypted_nfs or qcow2_vol):
                    mock_conn.assert_called_once_with(
                        root_helper, fake_ip,
                        multipath_supported, enforce_multipath,
                        host=fake_host)
                    fake_connector.connect_volume.assert_called_once_with(
                        mock.ANY)
                    fake_connector.disconnect_volume.assert_called_once_with(
                        mock.ANY, fake_devinfo, force=True)
                    fake_conn_obj.assert_called_once_with(
                        mock.ANY, root_helper, conn=mock.ANY,
                        use_multipath=multipath_supported)
                    attach_create.assert_called_once_with(
                        fake_client, fake_volume.id, mode=attach_mode)
                    attach_update.assert_called_once_with(
                        fake_client, fake_attachment_id,
                        fake_conn_info, mountpoint='glance_store')
                    attach_complete.assert_called_once_with(
                        fake_client, fake_attachment_id)
                    attach_delete.assert_called_once_with(fake_client,
                                                          fake_attachment_id)
                else:
                    mock_conn.assert_called_once_with(
                        root_helper, fake_ip,
                        multipath_supported, enforce_multipath,
                        host=fake_host)
                    fake_connector.connect_volume.assert_not_called()
                    fake_connector.disconnect_volume.assert_not_called()
                    attach_create.assert_called_once_with(
                        fake_client, fake_volume.id, mode=attach_mode)
                    attach_update.assert_called_once_with(
                        fake_client, fake_attachment_id,
                        fake_conn_info, mountpoint='glance_store')
                    attach_delete.assert_called_once_with(
                        fake_client, fake_attachment_id)

    def test_open_cinder_volume_rw(self):
        self._test_open_cinder_volume('wb', 'rw', None)

    def test_open_cinder_volume_ro(self):
        self._test_open_cinder_volume('rb', 'ro', None)

    def test_open_cinder_volume_update_attachment_error(self):
        err = Exception("update attachment fake error")
        self._test_open_cinder_volume('rb', 'ro', None,
                                      update_attachment_error=err)

    def test_open_cinder_volume_nfs_encrypted(self):
        self._test_open_cinder_volume('rb', 'ro', None, encrypted_nfs=True)

    def test_open_cinder_volume_nfs_qcow2_volume(self):
        self._test_open_cinder_volume('rb', 'ro', None, qcow2_vol=True)

    def test_open_cinder_volume_multiattach_volume(self):
        self._test_open_cinder_volume('rb', 'ro', None, multiattach=True)

    def _fake_volume_type_check(self, name):
        if name != 'some_type':
            raise cinder.cinder_exception.NotFound(code=404)

    def _test_configure_add_valid_type(self):

        with mock.patch.object(self.store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(
                volume_types=mock.MagicMock(
                    find=self._fake_volume_type_check))
            # If volume type exists, no exception is raised
            self.store.configure_add()

    def _test_configure_add_invalid_type(self):

        with mock.patch.object(self.store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(
                volume_types=mock.MagicMock(
                    find=self._fake_volume_type_check))
            with mock.patch.object(cinder, 'LOG') as mock_log:
                self.store.configure_add()
                mock_log.warning.assert_called_with(
                    "Invalid `cinder_volume_type some_random_type`")

    def _get_uri_loc(self, fake_volume_uuid, is_multi_store=False):
        if is_multi_store:
            uri = "cinder://cinder1/%s" % fake_volume_uuid
            loc = location.get_location_from_uri_and_backend(
                uri, "cinder1", conf=self.conf)
        else:
            uri = "cinder://%s" % fake_volume_uuid
            loc = location.get_location_from_uri(uri, conf=self.conf)

        return loc

    def _test_cinder_get(self, is_multi_store=False):
        expected_size = 5 * units.Ki
        expected_file_contents = b"*" * expected_size
        volume_file = io.BytesIO(expected_file_contents)
        fake_client = mock.MagicMock(auth_token=None, management_url=None)
        fake_volume_uuid = str(uuid.uuid4())
        fake_volume = mock.MagicMock(id=fake_volume_uuid,
                                     metadata={'image_size': expected_size},
                                     status='available')
        fake_volume.manager.get.return_value = fake_volume
        fake_volumes = mock.MagicMock(get=lambda id: fake_volume)

        @contextlib.contextmanager
        def fake_open(client, volume, mode):
            self.assertEqual('rb', mode)
            yield volume_file

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc, \
                mock.patch.object(self.store, '_open_cinder_volume',
                                  side_effect=fake_open):
            mock_cc.return_value = mock.MagicMock(client=fake_client,
                                                  volumes=fake_volumes)

            loc = self._get_uri_loc(fake_volume_uuid,
                                    is_multi_store=is_multi_store)

            (image_file, image_size) = self.store.get(loc,
                                                      context=self.context)

            expected_num_chunks = 2
            data = b""
            num_chunks = 0

            for chunk in image_file:
                num_chunks += 1
                data += chunk
            self.assertEqual(expected_num_chunks, num_chunks)
            self.assertEqual(expected_file_contents, data)

    def _test_cinder_volume_not_found(self, method_call, mock_method):
        fake_volume_uuid = str(uuid.uuid4())
        loc = mock.MagicMock(volume_id=fake_volume_uuid)
        mock_not_found = {mock_method: mock.MagicMock(
            side_effect=cinder.cinder_exception.NotFound(code=404))}
        fake_volumes = mock.MagicMock(**mock_not_found)

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(volumes=fake_volumes)
            self.assertRaises(exceptions.NotFound, method_call, loc,
                              context=self.context)

    def test_cinder_get_volume_not_found(self):
        self._test_cinder_volume_not_found(self.store.get, 'get')

    def test_cinder_get_size_volume_not_found(self):
        self._test_cinder_volume_not_found(self.store.get_size, 'get')

    def test_cinder_delete_volume_not_found(self):
        self._test_cinder_volume_not_found(self.store.delete, 'delete')

    def test_cinder_get_client_exception(self):
        fake_volume_uuid = str(uuid.uuid4())
        loc = mock.MagicMock(volume_id=fake_volume_uuid)

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc:
            mock_cc.side_effect = (
                cinder.cinder_exception.ClientException(code=500))
            self.assertRaises(exceptions.BackendException, self.store.get, loc,
                              context=self.context)

    def _test_cinder_get_size(self, is_multi_store=False):
        fake_client = mock.MagicMock(auth_token=None, management_url=None)
        fake_volume_uuid = str(uuid.uuid4())
        fake_volume = mock.MagicMock(size=5, metadata={})
        fake_volumes = mock.MagicMock(get=lambda fake_volume_uuid: fake_volume)

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(client=fake_client,
                                                    volumes=fake_volumes)

            loc = self._get_uri_loc(fake_volume_uuid,
                                    is_multi_store=is_multi_store)

            image_size = self.store.get_size(loc, context=self.context)
            self.assertEqual(fake_volume.size * units.Gi, image_size)

    def _test_cinder_get_size_with_metadata(self, is_multi_store=False):
        fake_client = mock.MagicMock(auth_token=None, management_url=None)
        fake_volume_uuid = str(uuid.uuid4())
        expected_image_size = 4500 * units.Mi
        fake_volume = mock.MagicMock(
            size=5, metadata={'image_size': expected_image_size})
        fake_volumes = {fake_volume_uuid: fake_volume}

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(client=fake_client,
                                                    volumes=fake_volumes)

            loc = self._get_uri_loc(fake_volume_uuid,
                                    is_multi_store=is_multi_store)

            image_size = self.store.get_size(loc, context=self.context)
            self.assertEqual(expected_image_size, image_size)

    def test_cinder_get_size_generic_exception(self):
        fake_volume_uuid = str(uuid.uuid4())
        loc = mock.MagicMock(volume_id=fake_volume_uuid)
        fake_volumes = mock.MagicMock(
            get=mock.MagicMock(side_effect=Exception()))

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(volumes=fake_volumes)
            image_size = self.store.get_size(loc, context=self.context)
            self.assertEqual(0, image_size)

    def _test_cinder_add(self, fake_volume, volume_file, size_kb=5,
                         verifier=None, backend='glance_store',
                         is_multi_store=False):
        expected_image_id = str(uuid.uuid4())
        expected_size = size_kb * units.Ki
        expected_file_contents = b"*" * expected_size
        image_file = io.BytesIO(expected_file_contents)
        expected_checksum = md5(expected_file_contents,
                                usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_file_contents).hexdigest()

        expected_location = 'cinder://%s' % fake_volume.id
        if is_multi_store:
            # Default backend is 'glance_store' for single store but in case
            # of multi store, if the backend option is not passed, we should
            # assign it to the default i.e. 'cinder1'
            if backend == 'glance_store':
                backend = 'cinder1'
            expected_location = 'cinder://%s/%s' % (backend, fake_volume.id)
        self.config(cinder_volume_type='some_type', group=backend)

        fake_client = mock.MagicMock(auth_token=None, management_url=None)
        fake_volume.manager.get.return_value = fake_volume
        fake_volumes = mock.MagicMock(create=mock.Mock(
            return_value=fake_volume))

        @contextlib.contextmanager
        def fake_open(client, volume, mode):
            self.assertEqual('wb', mode)
            yield volume_file

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc, \
                mock.patch.object(self.store, '_open_cinder_volume',
                                  side_effect=fake_open):
            mock_cc.return_value = mock.MagicMock(client=fake_client,
                                                  volumes=fake_volumes)
            loc, size, checksum, multihash, metadata = self.store.add(
                expected_image_id, image_file, expected_size, self.hash_algo,
                self.context, verifier)
            self.assertEqual(expected_location, loc)
            self.assertEqual(expected_size, size)
            self.assertEqual(expected_checksum, checksum)
            self.assertEqual(expected_multihash, multihash)
            fake_volumes.create.assert_called_once_with(
                1,
                name='image-%s' % expected_image_id,
                metadata={'image_owner': self.context.project_id,
                          'glance_image_id': expected_image_id,
                          'image_size': str(expected_size)},
                volume_type='some_type')
            if is_multi_store:
                self.assertEqual(backend, metadata["store"])

    def test_cinder_add_volume_not_found(self):
        image_file = mock.MagicMock()
        fake_image_id = str(uuid.uuid4())
        expected_size = 0
        fake_volumes = mock.MagicMock(create=mock.MagicMock(
            side_effect=cinder.cinder_exception.NotFound(code=404)))

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc:
            mock_cc.return_value = mock.MagicMock(volumes=fake_volumes)
            self.assertRaises(
                exceptions.BackendException, self.store.add,
                fake_image_id, image_file, expected_size, self.hash_algo,
                self.context, None)

    def _test_cinder_add_extend(self, is_multi_store=False, online=False):

        expected_volume_size = 2 * units.Gi
        expected_multihash = 'fake_hash'

        fakebuffer = mock.MagicMock()

        # CPython implementation detail: __len__ cannot return > sys.maxsize,
        # which on a 32-bit system is 2*units.Gi - 1
        # https://docs.python.org/3/reference/datamodel.html#object.__len__
        fakebuffer.__len__.return_value = int(expected_volume_size / 2)

        def get_fake_hash(type, secure=False):
            if type == 'md5':
                return mock.MagicMock(hexdigest=lambda: expected_checksum)
            else:
                return mock.MagicMock(hexdigest=lambda: expected_multihash)

        expected_image_id = str(uuid.uuid4())
        expected_volume_id = str(uuid.uuid4())
        expected_size = 0
        image_file = mock.MagicMock(
            read=mock.MagicMock(side_effect=[fakebuffer, fakebuffer, None]))
        fake_volume = mock.MagicMock(id=expected_volume_id, status='available',
                                     size=1)
        expected_checksum = 'fake_checksum'
        verifier = None
        backend = 'glance_store'

        expected_location = 'cinder://%s' % fake_volume.id
        if is_multi_store:
            # Default backend is 'glance_store' for single store but in case
            # of multi store, if the backend option is not passed, we should
            # assign it to the default i.e. 'cinder1'
            backend = 'cinder1'
            expected_location = 'cinder://%s/%s' % (backend, fake_volume.id)
        self.config(cinder_volume_type='some_type', group=backend)
        if online:
            self.config(cinder_do_extend_attached=True, group=backend)
            fake_connector = mock.MagicMock()
            fake_vol_connector_map = {expected_volume_id: fake_connector}
            self.store.volume_connector_map = fake_vol_connector_map

        fake_client = mock.MagicMock(auth_token=None, management_url=None)
        fake_volume.manager.get.return_value = fake_volume
        fake_volumes = mock.MagicMock(create=mock.Mock(
            return_value=fake_volume))

        @contextlib.contextmanager
        def fake_open(client, volume, mode):
            self.assertEqual('wb', mode)
            yield mock.MagicMock()

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc, \
                mock.patch.object(self.store, '_open_cinder_volume',
                                  side_effect=fake_open), \
                mock.patch.object(cinder.utils, 'get_hasher') as fake_hasher, \
                mock.patch.object(cinder.Store, '_wait_volume_status',
                                  return_value=fake_volume) as mock_wait, \
                mock.patch.object(cinder_utils.API,
                                  'extend_volume') as extend_vol:
            mock_cc_return_val = mock.MagicMock(client=fake_client,
                                                volumes=fake_volumes)
            mock_cc.return_value = mock_cc_return_val

            fake_hasher.side_effect = get_fake_hash
            loc, size, checksum, multihash, metadata = self.store.add(
                expected_image_id, image_file, expected_size, self.hash_algo,
                self.context, verifier)
            self.assertEqual(expected_location, loc)
            self.assertEqual(expected_volume_size, size)
            self.assertEqual(expected_checksum, checksum)
            self.assertEqual(expected_multihash, multihash)
            fake_volumes.create.assert_called_once_with(
                1,
                name='image-%s' % expected_image_id,
                metadata={'image_owner': self.context.project_id,
                          'glance_image_id': expected_image_id,
                          'image_size': str(expected_volume_size)},
                volume_type='some_type')
            if is_multi_store:
                self.assertEqual(backend, metadata["store"])
            if online:
                extend_vol.assert_called_once_with(
                    mock_cc_return_val, fake_volume,
                    expected_volume_size // units.Gi)
                mock_wait.assert_has_calls(
                    [mock.call(fake_volume, 'creating', 'available'),
                     mock.call(fake_volume, 'extending', 'in-use')])
            else:
                fake_volume.extend.assert_called_once_with(
                    fake_volume, expected_volume_size // units.Gi)
                mock_wait.assert_has_calls(
                    [mock.call(fake_volume, 'creating', 'available'),
                     mock.call(fake_volume, 'extending', 'available')])

    def test_cinder_add_extend_storage_full(self):

        expected_volume_size = 2 * units.Gi

        fakebuffer = mock.MagicMock()

        # CPython implementation detail: __len__ cannot return > sys.maxsize,
        # which on a 32-bit system is 2*units.Gi - 1
        # https://docs.python.org/3/reference/datamodel.html#object.__len__
        fakebuffer.__len__.return_value = int(expected_volume_size / 2)

        expected_image_id = str(uuid.uuid4())
        expected_volume_id = str(uuid.uuid4())
        expected_size = 0
        image_file = mock.MagicMock(
            read=mock.MagicMock(side_effect=[fakebuffer, fakebuffer, None]))
        fake_volume = mock.MagicMock(id=expected_volume_id, status='available',
                                     size=1)
        verifier = None

        fake_client = mock.MagicMock()
        fake_volume.manager.get.return_value = fake_volume
        fake_volumes = mock.MagicMock(create=mock.Mock(
            return_value=fake_volume))

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc, \
                mock.patch.object(self.store, '_open_cinder_volume'), \
                mock.patch.object(cinder.utils, 'get_hasher'), \
                mock.patch.object(
                    cinder.Store, '_wait_volume_status') as mock_wait:

            mock_cc.return_value = mock.MagicMock(client=fake_client,
                                                  volumes=fake_volumes)

            mock_wait.side_effect = [fake_volume, exceptions.BackendException]
            self.assertRaises(
                exceptions.StorageFull, self.store.add, expected_image_id,
                image_file, expected_size, self.hash_algo, self.context,
                verifier)

    def test_cinder_add_extend_volume_delete_exception(self):

        expected_volume_size = 2 * units.Gi

        fakebuffer = mock.MagicMock()

        # CPython implementation detail: __len__ cannot return > sys.maxsize,
        # which on a 32-bit system is 2*units.Gi - 1
        # https://docs.python.org/3/reference/datamodel.html#object.__len__
        fakebuffer.__len__.return_value = int(expected_volume_size / 2)

        expected_image_id = str(uuid.uuid4())
        expected_volume_id = str(uuid.uuid4())
        expected_size = 0
        image_file = mock.MagicMock(
            read=mock.MagicMock(side_effect=[fakebuffer, fakebuffer, None]))
        fake_volume = mock.MagicMock(
            id=expected_volume_id, status='available', size=1,
            delete=mock.MagicMock(side_effect=Exception()))

        fake_client = mock.MagicMock()
        fake_volume.manager.get.return_value = fake_volume
        fake_volumes = mock.MagicMock(create=mock.Mock(
            return_value=fake_volume))
        verifier = None

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc, \
                mock.patch.object(self.store, '_open_cinder_volume'), \
                mock.patch.object(cinder.utils, 'get_hasher'), \
                mock.patch.object(
                    cinder.Store, '_wait_volume_status') as mock_wait:

            mock_cc.return_value = mock.MagicMock(client=fake_client,
                                                  volumes=fake_volumes)

            mock_wait.side_effect = [fake_volume, exceptions.BackendException]
            self.assertRaises(
                exceptions.StorageFull, self.store.add, expected_image_id,
                image_file, expected_size, self.hash_algo, self.context,
                verifier)
            fake_volume.delete.assert_called_once()

    def _test_cinder_delete(self, is_multi_store=False):
        fake_client = mock.MagicMock(auth_token=None, management_url=None)
        fake_volume_uuid = str(uuid.uuid4())
        fake_volumes = mock.MagicMock(delete=mock.Mock())

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(client=fake_client,
                                                    volumes=fake_volumes)

            loc = self._get_uri_loc(fake_volume_uuid,
                                    is_multi_store=is_multi_store)

            self.store.delete(loc, context=self.context)
            fake_volumes.delete.assert_called_once_with(fake_volume_uuid)

    def test_cinder_delete_client_exception(self):
        fake_volume_uuid = str(uuid.uuid4())
        loc = mock.MagicMock(volume_id=fake_volume_uuid)
        fake_volumes = mock.MagicMock(delete=mock.MagicMock(
            side_effect=cinder.cinder_exception.ClientException(code=500)))

        with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
            mocked_cc.return_value = mock.MagicMock(volumes=fake_volumes)
            self.assertRaises(exceptions.BackendException, self.store.delete,
                              loc, context=self.context)

    def test_process_specs(self):
        self.location.process_specs()
        self.assertEqual('cinder', self.location.scheme)
        self.assertEqual(self.volume_id, self.location.volume_id)

    def _test_get_uri(self, expected_uri):
        uri = self.location.get_uri()
        self.assertEqual(expected_uri, uri)

    def _test_parse_uri_invalid(self, uri):
        self.assertRaises(
            exceptions.BadStoreUri, self.location.parse_uri, uri)

    def _test_get_root_helper(self, group='glance_store'):
        fake_rootwrap = 'fake_rootwrap'
        expected = 'sudo glance-rootwrap %s' % fake_rootwrap
        self.config(rootwrap_config=fake_rootwrap, group=group)
        res = self.store.get_root_helper()
        self.assertEqual(expected, res)

    def test_get_hash_str(self):
        nfs_conn = nfs.NfsBrickConnector()
        test_str = 'test_str'
        with mock.patch.object(nfs.hashlib, 'sha256') as fake_hashlib:
            nfs_conn.get_hash_str(test_str)
            test_str = test_str.encode('utf-8')
            fake_hashlib.assert_called_once_with(test_str)

    def test__get_mount_path(self):
        nfs_conn = nfs.NfsBrickConnector(mountpoint_base='fake_mount_path')
        fake_hex = 'fake_hex_digest'
        fake_share = 'fake_share'
        fake_path = 'fake_mount_path'
        expected_path = os.path.join(fake_path, fake_hex)
        with mock.patch.object(
                nfs.NfsBrickConnector, 'get_hash_str') as fake_hash:
            fake_hash.return_value = fake_hex
            res = nfs_conn._get_mount_path(fake_share, fake_path)
            self.assertEqual(expected_path, res)

    def test__get_host_ip_v6(self):
        fake_ipv6 = '2001:0db8:85a3:0000:0000:8a2e:0370'
        fake_socket_return = [[0, 1, 2, 3, [fake_ipv6]]]
        with mock.patch.object(cinder.socket, 'getaddrinfo') as fake_socket:
            fake_socket.return_value = fake_socket_return
            res = self.store._get_host_ip('fake_host')
            self.assertEqual(fake_ipv6, res)

    def test__get_host_ip_v4(self):
        fake_ip = '127.0.0.1'
        fake_socket_return = [[0, 1, 2, 3, [fake_ip]]]
        with mock.patch.object(cinder.socket, 'getaddrinfo') as fake_socket:
            fake_socket.side_effect = [socket.gaierror, fake_socket_return]
            res = self.store._get_host_ip('fake_host')
            self.assertEqual(fake_ip, res)



# Copyright 2021 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import contextlib
import logging
import socket
import threading

from oslo_config import cfg

from glance_store.common import cinder_utils
from glance_store import exceptions
from glance_store.i18n import _LE, _LW


LOG = logging.getLogger(__name__)

HOST = socket.gethostname()
CONF = cfg.CONF


class AttachmentStateManagerMeta(type):
    _instance = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instance:
            cls._instance[cls] = super(
                AttachmentStateManagerMeta, cls).__call__(*args, **kwargs)
        return cls._instance[cls]


class _AttachmentStateManager(metaclass=AttachmentStateManagerMeta):
    """A global manager of a volume's multiple attachments.

    _AttachmentStateManager manages a _AttachmentState object for the current
    glance node. Primarily it creates one on object initialization and returns
    it via get_state().

    _AttachmentStateManager manages concurrency itself. Independent callers do
    not need to consider interactions between multiple _AttachmentStateManager
    calls when designing their own locking.

    """
    # Reset state of global _AttachmentStateManager
    state = None
    use_count = 0

    # Guards both state and use_count
    cond = threading.Condition()

    def __init__(self, host):
        """Initialise a new _AttachmentState

        We will block before creating a new state until all operations
        using a previous state have completed.

        :param host: host
        """
        # Wait until all operations using a previous state are
        # complete before initialising a new one. Note that self.state is
        # already None, set either by initialisation or by host_down. This
        # means the current state will not be returned to any new callers,
        # and use_count will eventually reach zero.
        # We do this to avoid a race between _AttachmentState initialisation
        # and an on-going attach/detach operation
        self.host = host
        while self.use_count != 0:
            self.cond.wait()

        # Another thread might have initialised state while we were
        # waiting
        if self.state is None:
            LOG.debug('Initialising _AttachmentStateManager')
            self.state = _AttachmentState()

    @contextlib.contextmanager
    def get_state(self):
        """Return the current attachment state.

        _AttachmentStateManager will not permit a new state object to be
        created while any previous state object is still in use.

        :rtype: _AttachmentState
        """

        # We hold the instance lock here so that if a _AttachmentState is
        # currently initialising we'll wait for it to complete rather than
        # fail.
        with self.cond:
            state = self.state
            if state is None:
                LOG.error('Host not initialized')
                raise exceptions.HostNotInitialized(host=self.host)
            self.use_count += 1
        try:
            LOG.debug('Got _AttachmentState')
            yield state
        finally:
            with self.cond:
                self.use_count -= 1
                self.cond.notify_all()


class _AttachmentState(object):
    """A data structure recording all managed attachments. _AttachmentState
    ensures that the glance node only attempts to a single multiattach volume
    in use by multiple attachments once, and that it is not disconnected until
    it is no longer in use by any attachments.

    Callers should not create a _AttachmentState directly, but should obtain
    it via:

      with attachment.get_manager().get_state() as state:
        state.attach(...)

    _AttachmentState manages concurrency itself. Independent callers do not
    need to consider interactions between multiple _AttachmentState calls when
    designing their own locking.
    """

    class _Attachment(object):
        # A single multiattach volume, and the set of attachments in use
        # on it.
        def __init__(self):
            # A guard for operations on this volume
            self.lock = threading.Lock()

            # The set of attachments on this volume
            self.attachments = set()

        def add_attachment(self, attachment_id, host):
            self.attachments.add((attachment_id, host))

        def remove_attachment(self, attachment_id, host):
            self.attachments.remove((attachment_id, host))

        def in_use(self):
            return len(self.attachments) > 0

    def __init__(self):
        """Initialise _AttachmentState"""

        self.volumes = collections.defaultdict(self._Attachment)
        self.volume_api = cinder_utils.API()

    @contextlib.contextmanager
    def _get_locked(self, volume):
        """Get a locked attachment object

        :param mountpoint: The path of the volume whose attachment we should
                           return.
        :rtype: _AttachmentState._Attachment
        """
        while True:
            vol = self.volumes[volume]
            with vol.lock:
                if self.volumes[volume] is vol:
                    yield vol
                    break

    def attach(self, client, volume_id, host, mode=None):
        """Ensure a volume is available for an attachment and create an
        attachment

        :param client: Cinderclient object
        :param volume_id: ID of the volume to attach
        :param host: The host the volume will be attached to
        :param mode: The attachment mode
        """

        LOG.debug('_AttachmentState.attach(volume_id=%(volume_id)s, '
                  'host=%(host)s, mode=%(mode)s)',
                  {'volume_id': volume_id, 'host': host, 'mode': mode})
        with self._get_locked(volume_id) as vol_attachment:

            try:
                attachment = self.volume_api.attachment_create(
                    client, volume_id, mode=mode)
            except Exception:
                LOG.exception(_LE('Error attaching volume %(volume_id)s'),
                              {'volume_id': volume_id})
                del self.volumes[volume_id]
                raise

            vol_attachment.add_attachment(attachment['id'], host)

        LOG.debug('_AttachmentState.attach for volume_id=%(volume_id)s '
                  'and attachment_id=%(attachment_id)s completed successfully',
                  {'volume_id': volume_id, 'attachment_id': attachment['id']})
        return attachment

    def detach(self, client, attachment_id, volume_id, host, conn,
               connection_info, device):
        """Delete the attachment no longer in use, and disconnect volume
        if necessary.

        :param client: Cinderclient object
        :param attachment_id: ID of the attachment between volume and host
        :param volume_id: ID of the volume to attach
        :param host: The host the volume was attached to
        :param conn: connector object
        :param connection_info: connection information of the volume we are
                                detaching
        :device: device used to write image

        """
        LOG.debug('_AttachmentState.detach(vol_id=%(volume_id)s, '
                  'attachment_id=%(attachment_id)s)',
                  {'volume_id': volume_id, 'attachment_id': attachment_id})
        with self._get_locked(volume_id) as vol_attachment:
            try:
                vol_attachment.remove_attachment(attachment_id, host)
            except KeyError:
                LOG.warning(_LW("Request to remove attachment "
                                "(%(volume_id)s, %(host)s) but we "
                                "don't think it's in use."),
                            {'volume_id': volume_id, 'host': host})

            if not vol_attachment.in_use():
                conn.disconnect_volume(device)
                del self.volumes[volume_id]
            self.volume_api.attachment_delete(client, attachment_id)

            LOG.debug('_AttachmentState.detach for volume %(volume_id)s '
                      'and attachment_id=%(attachment_id)s completed '
                      'successfully',
                      {'volume_id': volume_id,
                       'attachment_id': attachment_id})


__manager__ = _AttachmentStateManager(HOST)


def attach(client, volume_id, host, mode=None):
    """A convenience wrapper around _AttachmentState.attach()"""

    with __manager__.get_state() as attach_state:
        attachment = attach_state.attach(client, volume_id, host, mode=mode)
    return attachment


def detach(client, attachment_id, volume_id, host, conn, connection_info,
           device):
    """A convenience wrapper around _AttachmentState.detach()"""

    with __manager__.get_state() as attach_state:
        attach_state.detach(client, attachment_id, volume_id, host, conn,
                            connection_info, device)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from glance_store import capabilities as caps
from glance_store.tests import base


class FakeStoreWithStaticCapabilities(caps.StoreCapability):
    _CAPABILITIES = caps.BitMasks.READ_RANDOM | caps.BitMasks.DRIVER_REUSABLE


class FakeStoreWithDynamicCapabilities(caps.StoreCapability):
    def __init__(self, *cap_list):
        super(FakeStoreWithDynamicCapabilities, self).__init__()
        if not cap_list:
            cap_list = [caps.BitMasks.READ_RANDOM,
                        caps.BitMasks.DRIVER_REUSABLE]
        self.set_capabilities(*cap_list)


class FakeStoreWithMixedCapabilities(caps.StoreCapability):
    _CAPABILITIES = caps.BitMasks.READ_RANDOM

    def __init__(self):
        super(FakeStoreWithMixedCapabilities, self).__init__()
        self.set_capabilities(caps.BitMasks.DRIVER_REUSABLE)


class TestStoreCapabilitiesChecking(object):

    def test_store_capabilities_checked_on_io_operations(self):
        self.assertEqual('op_checker', self.store.add.__name__)
        self.assertEqual('op_checker', self.store.get.__name__)
        self.assertEqual('op_checker', self.store.delete.__name__)


class TestStoreCapabilities(base.StoreBaseTest):

    def _verify_store_capabilities(self, store):
        # This function tested is_capable() as well.
        self.assertTrue(store.is_capable(caps.BitMasks.READ_RANDOM))
        self.assertTrue(store.is_capable(caps.BitMasks.DRIVER_REUSABLE))
        self.assertFalse(store.is_capable(caps.BitMasks.WRITE_ACCESS))

    def test_static_capabilities_setup(self):
        self._verify_store_capabilities(FakeStoreWithStaticCapabilities())

    def test_dynamic_capabilities_setup(self):
        self._verify_store_capabilities(FakeStoreWithDynamicCapabilities())

    def test_mixed_capabilities_setup(self):
        self._verify_store_capabilities(FakeStoreWithMixedCapabilities())

    def test_set_unset_capabilities(self):
        store = FakeStoreWithStaticCapabilities()
        self.assertFalse(store.is_capable(caps.BitMasks.WRITE_ACCESS))

        # Set and unset single capability on one time
        store.set_capabilities(caps.BitMasks.WRITE_ACCESS)
        self.assertTrue(store.is_capable(caps.BitMasks.WRITE_ACCESS))
        store.unset_capabilities(caps.BitMasks.WRITE_ACCESS)
        self.assertFalse(store.is_capable(caps.BitMasks.WRITE_ACCESS))

        # Set and unset multiple capabilities on one time
        cap_list = [caps.BitMasks.WRITE_ACCESS, caps.BitMasks.WRITE_OFFSET]
        store.set_capabilities(*cap_list)
        self.assertTrue(store.is_capable(*cap_list))
        store.unset_capabilities(*cap_list)
        self.assertFalse(store.is_capable(*cap_list))

    def test_store_capabilities_property(self):
        store1 = FakeStoreWithDynamicCapabilities()
        self.assertTrue(hasattr(store1, 'capabilities'))
        store2 = FakeStoreWithMixedCapabilities()
        self.assertEqual(store1.capabilities, store2.capabilities)

    def test_cascaded_unset_capabilities(self):
        # Test read capability
        store = FakeStoreWithMixedCapabilities()
        self._verify_store_capabilities(store)
        store.unset_capabilities(caps.BitMasks.READ_ACCESS)
        cap_list = [caps.BitMasks.READ_ACCESS, caps.BitMasks.READ_OFFSET,
                    caps.BitMasks.READ_CHUNK, caps.BitMasks.READ_RANDOM]
        for cap in cap_list:
            # To make sure all of them are unsetted.
            self.assertFalse(store.is_capable(cap))
        self.assertTrue(store.is_capable(caps.BitMasks.DRIVER_REUSABLE))

        # Test write capability
        store = FakeStoreWithDynamicCapabilities(caps.BitMasks.WRITE_RANDOM,
                                                 caps.BitMasks.DRIVER_REUSABLE)
        self.assertTrue(store.is_capable(caps.BitMasks.WRITE_RANDOM))
        self.assertTrue(store.is_capable(caps.BitMasks.DRIVER_REUSABLE))
        store.unset_capabilities(caps.BitMasks.WRITE_ACCESS)
        cap_list = [caps.BitMasks.WRITE_ACCESS, caps.BitMasks.WRITE_OFFSET,
                    caps.BitMasks.WRITE_CHUNK, caps.BitMasks.WRITE_RANDOM]
        for cap in cap_list:
            # To make sure all of them are unsetted.
            self.assertFalse(store.is_capable(cap))
        self.assertTrue(store.is_capable(caps.BitMasks.DRIVER_REUSABLE))


class TestStoreCapabilityConstants(base.StoreBaseTest):

    def test_one_single_capability_own_one_bit(self):
        cap_list = [
            caps.BitMasks.READ_ACCESS,
            caps.BitMasks.WRITE_ACCESS,
            caps.BitMasks.DRIVER_REUSABLE,
        ]
        for cap in cap_list:
            self.assertEqual(1, bin(cap).count('1'))

    def test_combined_capability_bits(self):
        check = caps.StoreCapability.contains
        check(caps.BitMasks.READ_OFFSET, caps.BitMasks.READ_ACCESS)
        check(caps.BitMasks.READ_CHUNK, caps.BitMasks.READ_ACCESS)
        check(caps.BitMasks.READ_RANDOM, caps.BitMasks.READ_CHUNK)
        check(caps.BitMasks.READ_RANDOM, caps.BitMasks.READ_OFFSET)
        check(caps.BitMasks.WRITE_OFFSET, caps.BitMasks.WRITE_ACCESS)
        check(caps.BitMasks.WRITE_CHUNK, caps.BitMasks.WRITE_ACCESS)
        check(caps.BitMasks.WRITE_RANDOM, caps.BitMasks.WRITE_CHUNK)
        check(caps.BitMasks.WRITE_RANDOM, caps.BitMasks.WRITE_OFFSET)
        check(caps.BitMasks.RW_ACCESS, caps.BitMasks.READ_ACCESS)
        check(caps.BitMasks.RW_ACCESS, caps.BitMasks.WRITE_ACCESS)
        check(caps.BitMasks.RW_OFFSET, caps.BitMasks.READ_OFFSET)
        check(caps.BitMasks.RW_OFFSET, caps.BitMasks.WRITE_OFFSET)
        check(caps.BitMasks.RW_CHUNK, caps.BitMasks.READ_CHUNK)
        check(caps.BitMasks.RW_CHUNK, caps.BitMasks.WRITE_CHUNK)
        check(caps.BitMasks.RW_RANDOM, caps.BitMasks.READ_RANDOM)
        check(caps.BitMasks.RW_RANDOM, caps.BitMasks.WRITE_RANDOM)



# Copyright 2010-2011 OpenStack Foundation
# Copyright 2018 Verizon Wireless
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import logging

from oslo_config import cfg
from oslo_utils import encodeutils
from stevedore import driver
from stevedore import extension

from glance_store import capabilities
from glance_store import exceptions
from glance_store.i18n import _
from glance_store import location

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

_STORE_OPTS = [
    cfg.ListOpt('stores',
                default=['file', 'http'],
                deprecated_for_removal=True,
                deprecated_since='Rocky',
                deprecated_reason="""
This option is deprecated against new config option
``enabled_backends`` which helps to configure multiple backend stores
of different schemes.

This option is scheduled for removal in the U development
cycle.
""",
                help="""
List of enabled Glance stores.

Register the storage backends to use for storing disk images
as a comma separated list. The default stores enabled for
storing disk images with Glance are ``file`` and ``http``.

Possible values:
    * A comma separated list that could include:
        * file
        * http
        * swift
        * rbd
        * cinder
        * vmware
        * s3

Related Options:
    * default_store

"""),
    cfg.StrOpt('default_store',
               default='file',
               choices=('file', 'filesystem', 'http', 'https', 'swift',
                        'swift+http', 'swift+https', 'swift+config', 'rbd',
                        'cinder', 'vsphere', 's3'),
               deprecated_for_removal=True,
               deprecated_since='Rocky',
               deprecated_reason="""
This option is deprecated against new config option
``default_backend`` which acts similar to ``default_store`` config
option.

This option is scheduled for removal in the U development
cycle.
""",
               help="""
The default scheme to use for storing images.

Provide a string value representing the default scheme to use for
storing images. If not set, Glance uses ``file`` as the default
scheme to store images with the ``file`` store.

NOTE: The value given for this configuration option must be a valid
scheme for a store registered with the ``stores`` configuration
option.

Possible values:
    * file
    * filesystem
    * http
    * https
    * swift
    * swift+http
    * swift+https
    * swift+config
    * rbd
    * cinder
    * vsphere
    * s3

Related Options:
    * stores

"""),
]

_STORE_CFG_GROUP = 'glance_store'


def _list_opts():
    driver_opts = []
    mgr = extension.ExtensionManager('glance_store.drivers')
    # NOTE(zhiyan): Handle available drivers entry_points provided
    # NOTE(nikhil): Return a sorted list of drivers to ensure that the sample
    # configuration files generated by oslo config generator retain the order
    # in which the config opts appear across different runs. If this order of
    # config opts is not preserved, some downstream packagers may see a long
    # diff of the changes though not relevant as only order has changed. See
    # some more details at bug 1619487.
    drivers = sorted([ext.name for ext in mgr])
    handled_drivers = []  # Used to handle backwards-compatible entries
    for store_entry in drivers:
        driver_cls = _load_store(None, store_entry, False)
        if driver_cls and driver_cls not in handled_drivers:
            if getattr(driver_cls, 'OPTIONS', None) is not None:
                driver_opts += driver_cls.OPTIONS
            handled_drivers.append(driver_cls)

    # NOTE(zhiyan): This separated approach could list
    # store options before all driver ones, which easier
    # to read and configure by operator.
    return ([(_STORE_CFG_GROUP, _STORE_OPTS)] +
            [(_STORE_CFG_GROUP, driver_opts)])


def register_opts(conf):
    opts = _list_opts()
    for group, opt_list in opts:
        LOG.debug("Registering options for group %s" % group)
        for opt in opt_list:
            conf.register_opt(opt, group=group)


class Indexable(object):
    """Indexable for file-like objs iterators

    Wrapper that allows an iterator or filelike be treated as an indexable
    data structure. This is required in the case where the return value from
    Store.get() is passed to Store.add() when adding a Copy-From image to a
    Store where the client library relies on eventlet GreenSockets, in which
    case the data to be written is indexed over.
    """

    def __init__(self, wrapped, size):
        """
        Initialize the object

        :param wrappped: the wrapped iterator or filelike.
        :param size: the size of data available
        """
        self.wrapped = wrapped
        self.size = int(size) if size else (wrapped.len
                                            if hasattr(wrapped, 'len') else 0)
        self.cursor = 0
        self.chunk = None

    def __iter__(self):
        """
        Delegate iteration to the wrapped instance.
        """
        for self.chunk in self.wrapped:
            yield self.chunk

    def __getitem__(self, i):
        """
        Index into the next chunk (or previous chunk in the case where
        the last data returned was not fully consumed).

        :param i: a slice-to-the-end
        """
        start = (i.start or 0) if isinstance(i, slice) else i
        if start < self.cursor:
            return self.chunk[(start - self.cursor):]

        self.chunk = self.another()
        if self.chunk:
            self.cursor += len(self.chunk)

        return self.chunk

    def another(self):
        """Implemented by subclasses to return the next element."""
        raise NotImplementedError

    def getvalue(self):
        """
        Return entire string value... used in testing
        """
        return self.wrapped.getvalue()

    def __len__(self):
        """
        Length accessor.
        """
        return self.size


def _load_store(conf, store_entry, invoke_load=True):
    try:
        LOG.debug("Attempting to import store %s", store_entry)
        mgr = driver.DriverManager('glance_store.drivers',
                                   store_entry,
                                   invoke_args=[conf],
                                   invoke_on_load=invoke_load)
        return mgr.driver
    except RuntimeError as e:
        LOG.warning("Failed to load driver %(driver)s. The "
                    "driver will be disabled" % dict(driver=str([driver, e])))


def _load_stores(conf):
    for store_entry in set(conf.glance_store.stores):
        try:
            # FIXME(flaper87): Don't hide BadStoreConfiguration
            # exceptions. These exceptions should be propagated
            # to the user of the library.
            store_instance = _load_store(conf, store_entry)

            if not store_instance:
                continue

            yield (store_entry, store_instance)

        except exceptions.BadStoreConfiguration:
            continue


def create_stores(conf=CONF):
    """
    Registers all store modules and all schemes
    from the given config. Duplicates are not re-registered.
    """
    store_count = 0

    for (store_entry, store_instance) in _load_stores(conf):
        try:
            schemes = store_instance.get_schemes()
            store_instance.configure(re_raise_bsc=False)
        except NotImplementedError:
            continue
        if not schemes:
            raise exceptions.BackendException('Unable to register store %s. '
                                              'No schemes associated with it.'
                                              % store_entry)
        else:
            LOG.debug("Registering store %s with schemes %s",
                      store_entry, schemes)

            scheme_map = {}
            loc_cls = store_instance.get_store_location_class()
            for scheme in schemes:
                scheme_map[scheme] = {
                    'store': store_instance,
                    'location_class': loc_cls,
                    'store_entry': store_entry
                }
            location.register_scheme_map(scheme_map)
            store_count += 1

    return store_count


def verify_default_store():
    scheme = CONF.glance_store.default_store
    try:
        get_store_from_scheme(scheme)
    except exceptions.UnknownScheme:
        msg = _("Store for scheme %s not found") % scheme
        raise RuntimeError(msg)


def get_known_schemes():
    """Returns list of known schemes."""
    return location.SCHEME_TO_CLS_MAP.keys()


def get_store_from_scheme(scheme):
    """
    Given a scheme, return the appropriate store object
    for handling that scheme.
    """
    if scheme not in location.SCHEME_TO_CLS_MAP:
        raise exceptions.UnknownScheme(scheme=scheme)
    scheme_info = location.SCHEME_TO_CLS_MAP[scheme]
    store = scheme_info['store']
    if not store.is_capable(capabilities.BitMasks.DRIVER_REUSABLE):
        # Driver instance isn't stateless so it can't
        # be reused safely and need recreation.
        store_entry = scheme_info['store_entry']
        store = _load_store(store.conf, store_entry, invoke_load=True)
        store.configure()
        try:
            scheme_map = {}
            loc_cls = store.get_store_location_class()
            for scheme in store.get_schemes():
                scheme_map[scheme] = {
                    'store': store,
                    'location_class': loc_cls,
                    'store_entry': store_entry
                }
                location.register_scheme_map(scheme_map)
        except NotImplementedError:
            scheme_info['store'] = store
    return store


def get_store_from_uri(uri):
    """
    Given a URI, return the store object that would handle
    operations on the URI.

    :param uri: URI to analyze
    """
    scheme = uri[0:uri.find('/') - 1]
    return get_store_from_scheme(scheme)


def get_from_backend(uri, offset=0, chunk_size=None, context=None):
    """Yields chunks of data from backend specified by uri."""

    loc = location.get_location_from_uri(uri, conf=CONF)
    store = get_store_from_uri(uri)

    return store.get(loc, offset=offset,
                     chunk_size=chunk_size,
                     context=context)


def get_size_from_backend(uri, context=None):
    """Retrieves image size from backend specified by uri."""

    loc = location.get_location_from_uri(uri, conf=CONF)
    store = get_store_from_uri(uri)
    return store.get_size(loc, context=context)


def delete_from_backend(uri, context=None):
    """Removes chunks of data from backend specified by uri."""

    loc = location.get_location_from_uri(uri, conf=CONF)
    store = get_store_from_uri(uri)
    return store.delete(loc, context=context)


def get_store_from_location(uri):
    """
    Given a location (assumed to be a URL), attempt to determine
    the store from the location.  We use here a simple guess that
    the scheme of the parsed URL is the store...

    :param uri: Location to check for the store
    """
    loc = location.get_location_from_uri(uri, conf=CONF)
    return loc.store_name


def check_location_metadata(val, key=''):
    if isinstance(val, dict):
        for key in val:
            check_location_metadata(val[key], key=key)
    elif isinstance(val, list):
        ndx = 0
        for v in val:
            check_location_metadata(v, key='%s[%d]' % (key, ndx))
            ndx = ndx + 1
    elif not isinstance(val, str):
        raise exceptions.BackendException(_("The image metadata key %(key)s "
                                            "has an invalid type of %(type)s. "
                                            "Only dict, list, and unicode are "
                                            "supported.")
                                          % dict(key=key, type=type(val)))


def _check_metadata(store, metadata):
    if not isinstance(metadata, dict):
        msg = (_("The storage driver %(driver)s returned invalid "
                 " metadata %(metadata)s. This must be a dictionary type")
               % dict(driver=str(store), metadata=str(metadata)))
        LOG.error(msg)
        raise exceptions.BackendException(msg)
    try:
        check_location_metadata(metadata)
    except exceptions.BackendException as e:
        e_msg = (_("A bad metadata structure was returned from the "
                   "%(driver)s storage driver: %(metadata)s.  %(e)s.") %
                 dict(driver=encodeutils.exception_to_unicode(store),
                      metadata=encodeutils.exception_to_unicode(metadata),
                      e=encodeutils.exception_to_unicode(e)))
        LOG.error(e_msg)
        raise exceptions.BackendException(e_msg)


def store_add_to_backend(image_id, data, size, store, context=None,
                         verifier=None):
    """
    A wrapper around a call to each stores add() method.  This gives glance
    a common place to check the output

    :param image_id:  The image add to which data is added
    :param data: The data to be stored
    :param size: The length of the data in bytes
    :param store: The store to which the data is being added
    :param context: The request context
    :param verifier: An object used to verify signatures for images
    :return: The url location of the file,
             the size amount of data,
             the checksum of the data
             the storage systems metadata dictionary for the location
    """
    (location, size, checksum, metadata) = store.add(image_id,
                                                     data,
                                                     size,
                                                     context=context,
                                                     verifier=verifier)
    if metadata is not None:
        _check_metadata(store, metadata)

    return (location, size, checksum, metadata)


def store_add_to_backend_with_multihash(
        image_id, data, size, hashing_algo, store,
        context=None, verifier=None):
    """
    A wrapper around a call to each store's add() method that requires
    a hashing_algo identifier and returns a 5-tuple including the
    "multihash" computed using the specified hashing_algo.  (This
    is an enhanced version of store_add_to_backend(), which is left
    as-is for backward compatibility.)

    :param image_id:  The image add to which data is added
    :param data: The data to be stored
    :param size: The length of the data in bytes
    :param store: The store to which the data is being added
    :param hashing_algo: A hashlib algorithm identifier (string)
    :param context: The request context
    :param verifier: An object used to verify signatures for images
    :return: The url location of the file,
             the size amount of data,
             the checksum of the data,
             the multihash of the data,
             the storage system's metadata dictionary for the location
    :raises: ``glance_store.exceptions.BackendException``
             ``glance_store.exceptions.UnknownHashingAlgo``
    """

    if hashing_algo not in hashlib.algorithms_available:
        raise exceptions.UnknownHashingAlgo(algo=hashing_algo)

    (location, size, checksum, multihash, metadata) = store.add(
        image_id, data, size, hashing_algo, context=context, verifier=verifier)

    if metadata is not None:
        _check_metadata(store, metadata)

    return (location, size, checksum, multihash, metadata)


def add_to_backend(conf, image_id, data, size, scheme=None, context=None,
                   verifier=None):
    if scheme is None:
        scheme = conf['glance_store']['default_store']
    store = get_store_from_scheme(scheme)
    return store_add_to_backend(image_id, data, size, store, context,
                                verifier)


def add_to_backend_with_multihash(conf, image_id, data, size, hashing_algo,
                                  scheme=None, context=None, verifier=None):
    if scheme is None:
        scheme = conf['glance_store']['default_store']
    store = get_store_from_scheme(scheme)
    return store_add_to_backend_with_multihash(
        image_id, data, size, hashing_algo, store, context, verifier)


def set_acls(location_uri, public=False, read_tenants=[],
             write_tenants=None, context=None):

    if write_tenants is None:
        write_tenants = []

    loc = location.get_location_from_uri(location_uri, conf=CONF)
    scheme = get_store_from_location(location_uri)
    store = get_store_from_scheme(scheme)
    try:
        store.set_acls(loc, public=public,
                       read_tenants=read_tenants,
                       write_tenants=write_tenants,
                       context=context)
    except NotImplementedError:
        LOG.debug(_("Skipping store.set_acls... not implemented."))



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['openstackdocstheme',
              'sphinxcontrib.apidoc']

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/glance_store'
openstackdocs_auto_name = False
openstackdocs_bug_project = 'glance-store'
openstackdocs_bug_tag = ''

# sphinxcontrib.apidoc options
apidoc_module_dir = '../../glance_store'
apidoc_output_dir = 'reference/api'
apidoc_excluded_paths = [
    'test',
    'tests/*']
apidoc_separate_modules = True

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# Add any paths that contain templates here, relative to this directory.
# templates_path = []

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'glance_store'
copyright = '2014, OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
# html_static_path = ['static']
html_theme = 'openstackdocs'

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project

modindex_common_prefix = ['glance_store.']

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     '%s.tex' % project,
     '%s Documentation' % project,
     'OpenStack Foundation', 'manual'),
]

# The autodoc module imports every module to check for import
# errors. Since the fs_mount module is self initializing, it
# requires configurations that aren't loaded till that time.
# It would never happen in a real scenario as it is only imported
# from cinder store after the config are loaded but to handle doc
# failures, we mock it here.
# The cinder_utils module imports external dependencies like
# cinderclient, retrying etc which are not recognized by
# autodoc, hence, are mocked here. These dependencies are installed
# during an actual deployment and won't cause any issue during usage.
autodoc_mock_imports = ['glance_store.common.fs_mount',
                        'glance_store.common.cinder_utils']

# Since version 4.2.0, Sphinx emits a warning when encountering a mocked
# object, leading to the following error:
#   "A mocked object is detected: 'glance_store.common.cinder_utils'"
# To prevent this, we disable all warnings from the autodoc extension, since
# there is no finer grain yet.
suppress_warnings = ['autodoc.*']



# Copyright 2010-2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Connection Manager for Swift connections that responsible for providing
connection with valid credentials and updated token"""

import logging

from oslo_utils import encodeutils

from glance_store import exceptions
from glance_store.i18n import _, _LI

LOG = logging.getLogger(__name__)


class SwiftConnectionManager(object):
    """Connection Manager class responsible for initializing and managing
    swiftclient connections in store. The instance of that class can provide
    swift connections with a valid(and refreshed) user token if the token is
    going to expire soon.
    """

    AUTH_HEADER_NAME = 'X-Auth-Token'

    def __init__(self, store, store_location, context=None,
                 allow_reauth=False):
        """Initialize manager with parameters required to establish connection.

        Initialize store and prepare it for interacting with swift. Also
        initialize keystone client that need to be used for authentication if
        allow_reauth is True.
        The method invariant is the following: if method was executed
        successfully and self.allow_reauth is True users can safely request
        valid(no expiration) swift connections any time. Otherwise, connection
        manager initialize a connection once and always returns that connection
        to users.

        :param store: store that provides connections
        :param store_location: image location in store
        :param context: user context to access data in Swift
        :param allow_reauth: defines if re-authentication need to be executed
        when a user request the connection
        """
        self._client = None
        self.store = store
        self.location = store_location
        self.context = context
        self.allow_reauth = allow_reauth
        self.storage_url = self._get_storage_url()
        self.connection = self._init_connection()

    def get_connection(self):
        """Get swift client connection.

        Returns swift client connection. If allow_reauth is True and
        connection token is going to expire soon then the method returns
        updated connection.
        The method invariant is the following: if self.allow_reauth is False
        then the method returns the same connection for every call. So the
        connection may expire. If self.allow_reauth is True the returned
        swift connection is always valid and cannot expire at least for
        swift_store_expire_soon_interval.
        """
        if self.allow_reauth:
            # we are refreshing token only and if only connection manager
            # re-authentication is allowed. Token refreshing is setup by
            # connection manager users. Also we disable re-authentication
            # if there is not way to execute it (cannot initialize trusts for
            # multi-tenant or auth_version is not 3)
            auth_ref = self.client.session.auth.auth_ref
            # if connection token is going to expire soon (keystone checks
            # is token is going to expire or expired already)
            if self.store.backend_group:
                interval = getattr(
                    self.store.conf, self.store.backend_group
                ).swift_store_expire_soon_interval
            else:
                store_conf = self.store.conf.glance_store
                interval = store_conf.swift_store_expire_soon_interval

            if auth_ref.will_expire_soon(interval):
                LOG.info(_LI("Requesting new token for swift connection."))
                # request new token with session and client provided by store
                auth_token = self.client.session.get_auth_headers().get(
                    self.AUTH_HEADER_NAME)
                LOG.info(_LI("Token has been successfully requested. "
                             "Refreshing swift connection."))
                # initialize new switclient connection with fresh token
                self.connection = self.store.get_store_connection(
                    auth_token, self.storage_url)
        return self.connection

    @property
    def client(self):
        """Return keystone client to request a  new token.

        Initialize a client lazily from the method provided by glance_store.
        The method invariant is the following: if client cannot be
        initialized raise exception otherwise return initialized client that
        can be used for re-authentication any time.
        """
        if self._client is None:
            self._client = self._init_client()
        return self._client

    def _init_connection(self):
        """Initialize and return valid Swift connection."""
        auth_token = self.client.session.get_auth_headers().get(
            self.AUTH_HEADER_NAME)
        return self.store.get_store_connection(
            auth_token, self.storage_url)

    def _init_client(self):
        """Initialize Keystone client."""
        return self.store.init_client(location=self.location,
                                      context=self.context)

    def _get_storage_url(self):
        """Request swift storage url."""
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class SingleTenantConnectionManager(SwiftConnectionManager):
    def _get_storage_url(self):
        """Get swift endpoint from keystone

        Return endpoint for swift from service catalog if not overridden in
        store configuration. The method works only Keystone v3.
        If you are using different version (1 or 2)
        it returns None.
        :return: swift endpoint
        """

        if self.store.conf_endpoint:
            return self.store.conf_endpoint

        if self.store.auth_version == '3':
            try:
                return self.client.session.get_endpoint(
                    service_type=self.store.service_type,
                    interface=self.store.endpoint_type,
                    region_name=self.store.region
                )
            except Exception as e:
                # do the same that swift driver does
                # when catching ClientException
                msg = _("Cannot find swift service endpoint : "
                        "%s") % encodeutils.exception_to_unicode(e)
                raise exceptions.BackendException(msg)

    def _init_connection(self):
        if self.store.auth_version == '3':
            return super(SingleTenantConnectionManager,
                         self)._init_connection()
        else:
            # no re-authentication for v1 and v2
            self.allow_reauth = False
            # use good old connection initialization
            return self.store.get_connection(self.location, self.context)


class MultiTenantConnectionManager(SwiftConnectionManager):

    def __init__(self, store, store_location, context=None,
                 allow_reauth=False):
        # no context - no party
        if context is None:
            reason = _("Multi-tenant Swift storage requires a user context.")
            raise exceptions.BadStoreConfiguration(store_name="swift",
                                                   reason=reason)
        super(MultiTenantConnectionManager, self).__init__(
            store, store_location, context, allow_reauth)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._client and self.client.trust_id:
            # client has been initialized - need to cleanup resources
            LOG.info(_LI("Revoking trust %s"), self.client.trust_id)
            self.client.trusts.delete(self.client.trust_id)

    def _get_storage_url(self):
        return self.location.swift_url

    def _init_connection(self):
        if self.allow_reauth:
            try:
                return super(MultiTenantConnectionManager,
                             self)._init_connection()
            except Exception as e:
                LOG.debug("Cannot initialize swift connection for multi-tenant"
                          " store with trustee token: %s. Using user token for"
                          " connection initialization.", e)
                # for multi-tenant store we have a token, so we can use it
                # for connection initialization but we cannot fetch new token
                # with client
                self.allow_reauth = False

        return self.store.get_store_connection(
            self.context.auth_token, self.storage_url)



# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import logging
import os
import socket

from oslo_config import cfg

from glance_store._drivers.cinder import base
from glance_store.common import cinder_utils
from glance_store.common import fs_mount as mount
from glance_store.common import utils
from glance_store import exceptions
from glance_store.i18n import _

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class NfsBrickConnector(base.BaseBrickConnectorInterface):

    def __init__(self, *args, **kwargs):
        self.volume = kwargs.get('volume')
        self.connection_info = kwargs.get('connection_info')
        self.root_helper = kwargs.get('root_helper')
        self.mount_point_base = kwargs.get('mountpoint_base')
        self.attachment_obj = kwargs.get('attachment_obj')
        self.client = kwargs.get('client')
        self.host = socket.gethostname()
        self.volume_api = cinder_utils.API()

    def _get_mount_path(self, share, mount_point_base):
        """Returns the mount path prefix using the mount point base and share.

        :returns: The mount path prefix.
        """
        return os.path.join(self.mount_point_base,
                            NfsBrickConnector.get_hash_str(share))

    @staticmethod
    def get_hash_str(base_str):
        """Returns string representing SHA256 hash of base_str in hex format.

        If base_str is a Unicode string, encode it to UTF-8.
        """
        if isinstance(base_str, str):
            base_str = base_str.encode('utf-8')
        return hashlib.sha256(base_str).hexdigest()

    def connect_volume(self, volume):
        # The format info of nfs volumes is exposed via attachment_get
        # API hence it is not available in the connection info of
        # attachment object received from attachment_update and we
        # need to do this call
        vol_attachment = self.volume_api.attachment_get(
            self.client, self.attachment_obj.id)
        if (volume.encrypted or
                vol_attachment.connection_info['format'] == 'qcow2'):
            issue_type = 'Encrypted' if volume.encrypted else 'qcow2'
            msg = (_('%(issue_type)s volume creation for cinder nfs '
                     'is not supported from glance_store. Failed to '
                     'create volume %(volume_id)s')
                   % {'issue_type': issue_type,
                      'volume_id': volume.id})
            LOG.error(msg)
            raise exceptions.BackendException(msg)

        @utils.synchronized(self.connection_info['export'])
        def connect_volume_nfs():
            export = self.connection_info['export']
            vol_name = self.connection_info['name']
            mountpoint = self._get_mount_path(
                export, os.path.join(self.mount_point_base, 'nfs'))
            options = self.connection_info['options']
            mount.mount(
                'nfs', export, vol_name, mountpoint, self.host,
                self.root_helper, options)
            return {'path': os.path.join(mountpoint, vol_name)}

        device = connect_volume_nfs()
        return device

    def disconnect_volume(self, device):
        @utils.synchronized(self.connection_info['export'])
        def disconnect_volume_nfs():
            path, vol_name = device['path'].rsplit('/', 1)
            mount.umount(vol_name, path, self.host,
                         self.root_helper)
        disconnect_volume_nfs()

    def extend_volume(self):
        raise NotImplementedError
